#include "plugin_manager.h"
#include "_aos.h"
#include "eef_aos.h"

static pthread_key_t           _aos_key;
static pthread_once_t  	       _aos_key_once = PTHREAD_ONCE_INIT; /* Normally not okay to do initialisation outside of a function, but this way it (hopefully) will only be set one - during compile time */
static struct aos_state_s *    _global_state;

static int                     _is_initialized;
static int                     _is_threading;

/* internal methods */

/**
 *  returns _is_initialized
 */
int aos_is_initialized(){
	return _is_initialized;
}

/**
 * sets threading flag
 */
void aos_start_threading(){
	_is_threading = 1;
}

/**
 * returns a pointer to an aos_storage_t struct that corresponds to the current thread context. If the current context is not threaded, a pointer to a global storage struct is returned.
 */
aos_state_t * aos_get_state(void){
  pid_t             tid;
  aos_state_t *     state = NULL;

  tid = syscall(SYS_gettid);
  /*tid = gettid();*/

  if(!_is_threading){
    return _global_state;
  }

  /*  Try to get TLS */
  state = pthread_getspecific(_aos_key);
  if(state == 0) {
    /* Create new TLS */
    if((state = calloc(1, sizeof(aos_state_t)))){
      state->last_context = NULL;
      if((state->current_storage = calloc(1, sizeof(aos_storage_t)))){
        state->current_storage->list_contexts = NULL;
        _is_initialized = 1;
        state->saved_storage = state->current_storage;
        pthread_setspecific(_aos_key, state);
        return state;
      }
    }
  } else {
    return state;
  }
  return NULL;
}

aos_storage_t* aos_get_storage(){
  aos_state_t* iterator = aos_get_state();
  return iterator->current_storage;
}

EES_RC aos_free_state(aos_state_t* state){
  /*printf("Freeing state at %p\n", state);*/
  aos_free_storage(state->saved_storage);
  state->saved_storage = NULL;
  state->last_context  = NULL;
  return EES_SUCCESS;
}

/**
 * Frees storage struct pointed to by storage pointer
 */
EES_RC aos_free_storage(aos_storage_t* storage){
  aos_context_t   *context        = NULL, *next_context    = NULL;
  aos_attribute_t *attribute      = NULL, *next_attribute  = NULL;
  int             attribute_count = 0;

  if(!storage){
    return EES_FAILURE;
  }

  EEF_log(LOG_DEBUG, "Cleaning aos storage at %p\n", storage);

  context = storage->list_contexts;

  while(context){
    next_context = context->next;
    attribute = context->list_attributes;
    while(attribute){
      next_attribute = attribute->next;
      free(attribute->id);
      free(attribute->issuer);
      /*free(attribute->type);*/
      if(attribute->needs_free){
        EEF_log(LOG_DEBUG, "Freeing arg data %i at %p\n",attribute_count, attribute->data);
        free(attribute->data);
      }
      free(attribute);
      attribute = next_attribute;
      attribute_count++;
    }
    /* free here */
    EEF_log(LOG_DEBUG, "Freeing context at %p\n", context);
    context->next = NULL;
    context->list_attributes = NULL;
    free(context);
    context = next_context;
  }
  storage->list_contexts = NULL;
  free(storage);
  storage = NULL;
  return EES_SUCCESS;
}

/* base functions */

/**
 * Returns a pointer to a freshly allocated aos_context_t struct
 */
aos_context_t * createContext(aos_context_class_t context_class){
  aos_context_t * context;
  if((context = calloc(1, sizeof(aos_context_t)))){
    context->context_class = context_class;
    context->list_attributes = NULL;
    context->last_attribute = NULL;
    context->next = NULL;
  }
  /*printf("Context of type %i created at %p\n", context_class, context);*/
  return context;
}

/**
 * Adds supplied context to storage attained through aos_get_storage()
 */
EES_RC addContext(aos_context_t* context){
  aos_state_t* iterator = aos_get_state();
  aos_storage_t* storage = iterator->saved_storage;
  aos_context_t* tmp_context = NULL;

  EEF_log(LOG_DEBUG, "Adding context at %p to storage at %p\n", context, iterator->current_storage);
  if(!storage){
    EEF_log(LOG_ERR, "No storage available at %p", storage);
    return EES_FAILURE;
  }
  if(storage->list_contexts != NULL){
    if((tmp_context = storage->list_contexts)!=NULL){
      while(tmp_context->next){
        if(tmp_context == context){
          EEF_log(LOG_ERR, "The context you're trying to add was already added!\n");
          return EES_FAILURE;
        }
        tmp_context = tmp_context->next;
      }
    }
    tmp_context->next = context;
    EEF_log(LOG_DEBUG, "Inserted at %p\n", tmp_context);
    /*printf("Inserted at %p\n", tmp_context);*/
  } else {
    storage->list_contexts = context;
    EEF_log(LOG_DEBUG, "Created at %p\n", context);
    /*printf("List created at %p\n", context);*/
  }
  aos_set_iterator(context);

  /*eef_log(LOG_DEBUG, "Copied context to %p\n", storage->last_context);*/
  return EES_SUCCESS;
}

EES_RC rewindContexts(aos_storage_t* storage){
  aos_state_t*    state = aos_get_state();
  /* not using storage parameter anymore */

  if(state){
    state->current_storage = state->saved_storage;

    if(state->current_storage){
      aos_set_iterator(state->current_storage->list_contexts);
      return EES_SUCCESS;
    } else {
      EEF_log(LOG_ERR, "Unable to aquire AOS handle in rewindContexts()\n");
    }
  }
  return EES_FAILURE;
}

EES_RC aos_set_iterator(aos_context_t* context){
  aos_state_t* state = aos_get_state();
  if(state->current_storage){
    /*printf("Setting iterator for %p to %p\n", state->current_storage, context);*/
    state->last_context = context;
    return EES_SUCCESS;
  } else {
    EEF_log(LOG_ERR, "Invalid storage passed to aos_set_iterator(): %p\n", state->current_storage);
  }
  return EES_FAILURE;
}

aos_context_t* aos_get_iterator(void){
  aos_state_t* state = aos_get_state();
  if(state->current_storage){
    /*printf("Getting iterator for %p as %p\n", state->current_storage, state->last_context);*/
    if(state->last_context){
      return state->last_context;
    }
  } else {
    EEF_log(LOG_ERR, "Invalid storage passed to aos_set_iterator(): %p\n", state->current_storage);
  }
  return NULL;
}

/**
 * returns the next aos_context_t pointer with the supplied context_class from the list of contexts in the supplied storage.
 * will fall through and try to use global storage in case nothing was found
 */
aos_context_t * getNextContext(aos_context_class_t context_class, aos_storage_t* storage){
  aos_state_t*    state           = aos_get_state();
  aos_context_t*  current_context = aos_get_iterator();

  /* find context type */
  if(context_class != ANY){
    while(current_context != NULL){
      if((current_context->context_class == context_class)){
        /* Found context */
        break;
      }
      current_context = current_context->next;
    }
  }

  /* save state */
  if((current_context == NULL) && (storage != _global_state->current_storage) && (state->current_storage != _global_state->current_storage)){
    /* Retry with global storage */
    state->current_storage = _global_state->current_storage;
    aos_set_iterator(_global_state->current_storage->list_contexts);
    current_context = getNextContext(context_class, _global_state->current_storage);
  }

  /* save iterator */
  if(current_context){
    aos_set_iterator(current_context->next);
  } else {
    aos_set_iterator(NULL);
  }

  return current_context;
}

/**
 * Sets the obligation name in the struct pointed to by the supplied pointer context to supplied obligation
 */
void setContextObligationId(aos_context_t* context, char * obligation){
  if(context->context_class == OBLIGATION){
    context->obligation_name = obligation;
  }
}

/**
 * Returns the obligation name from the supplied context
 */
char* getContextObligationId(aos_context_t* context){
  return context->obligation_name;
}

/**
 * Returns a pointer to a freshly allocated aos_attribute_t struct
 */
aos_attribute_t* createAttribute(void){
  aos_attribute_t* attribute = NULL;
  if((attribute = calloc(1, sizeof(aos_attribute_t)))){
    attribute->next = NULL;
    attribute->data = NULL;
    attribute->setting_plugin = get_running_plugin();
  }
  return attribute;
}

/**
 * Adds supplied attribute to supplied context
 */
EES_RC addAttribute(aos_context_t * context, aos_attribute_t* attribute){
  aos_attribute_t* working = NULL;
  if (context->list_attributes == NULL){
    context->list_attributes = attribute;
  } else {
    working = context->list_attributes;
    while(working->next)
      working = working->next;
    working->next = attribute;
  }
  return EES_SUCCESS;
}

EES_RC rewindAttributes(aos_context_t* context){
  context->last_attribute = NULL;
  return EES_SUCCESS;
}

/**
 * returns the next aos_attribute_t pointer from list of attributes in the supplied context
 */
aos_attribute_t* getNextAttribute(aos_context_t* context){
  aos_attribute_t* attribute = NULL;
  if(context->last_attribute){
    attribute = context->last_attribute->next;
  } else {
    attribute = context->list_attributes;
  }
  if (attribute)
    context->last_attribute = attribute;
  return attribute;
}

/**
 * removes a supplied attribute from the supplied context
 */
EES_RC destroyAttribute(aos_context_t* context, aos_attribute_t* attribute){
  aos_attribute_t *last_attribute = NULL, *current_attribute = NULL;
  for(current_attribute = context->list_attributes; current_attribute != NULL; last_attribute = current_attribute, current_attribute = current_attribute->next){
    if(current_attribute == attribute){
      if(current_attribute->setting_plugin != get_running_plugin()){
        EEF_log(LOG_ERR, "Argument %s is not owned by running plugin %s", current_attribute->id, get_running_plugin()->name);
        return EES_FAILURE;
      }
      free(attribute->data);
      free(attribute);
      attribute = NULL;
      if(last_attribute == NULL){
        context->list_attributes = current_attribute->next;
      } else {
        last_attribute->next = current_attribute->next;
      }
      break;
    }
  }

  if(current_attribute){
    free(current_attribute->data);
    free(current_attribute);
    current_attribute = NULL;
    return EES_SUCCESS;
  }
  return EES_FAILURE;
}

/**
 * sets attribute id in supplied attribute
 */
EES_RC setAttributeId(aos_attribute_t* attribute, char* id){
  free(attribute->id);
  attribute->id = NULL;
  if(id != NULL){
    attribute->id = strdup(id);
  } else {
    EEF_log(LOG_ERR, "setAttributeId(), tried to set NULL id\n");
  }
  return EES_SUCCESS;
}

/**
 * sets attribute issuer in supplied attribute
 */
EES_RC setAttributeIssuer(aos_attribute_t* attribute, char* issuer){
  free(attribute->issuer);
  attribute->issuer = NULL;
  if(issuer != NULL){
    attribute->issuer = strdup(issuer);
  } else {
    EEF_log(LOG_ERR, "setAttributeIssuer(), tried to set NULL issuer\n");
    return EES_FAILURE;
  }
  return EES_SUCCESS;
}

/**
 * sets attribute value in supplied attribute
 */
EES_RC setAttributeValue(aos_attribute_t* attribute, void* value, size_t size){
  if(value != NULL){
    if(attribute->needs_free == 1){
      free(attribute->data);
      attribute->data = NULL;
      attribute->needs_free = 0;
    }
    if(size == 0) {
      attribute->data = value;
      attribute->needs_free = 0;
    } else {
      attribute->data = calloc(1, size+1);
      memcpy(attribute->data, value, size);
      attribute->needs_free = 1;
    }
  } else {
    return EES_FAILURE;
  }
  return EES_SUCCESS;
}

/**
 * sets attribute type in supplied attribute
 */
EES_RC setAttributeType(aos_attribute_t * attribute, char* type){
  attribute->type = type;
  return EES_SUCCESS;
}

/**
 * returns type from supplied attribute
 */
char* getAttributeType(aos_attribute_t * attribute){
  return attribute->type;
}

/**
 * returns id from supplied attribute
 */
char* getAttributeId(aos_attribute_t* attribute){
  return attribute->id;
}

/**
 * returns issuer from supplied attribute
 */
char* getAttributeIssuer(aos_attribute_t* attribute){
  return attribute->issuer;
}

/**
 * returns data from supplied attribute casted to a char*
 */
char* getAttributeValueAsString(aos_attribute_t* attribute){
  return (char*) attribute->data;
}

/**
 * returns data from supplied attribute casted to a int
 */
int getAttributeValueAsInt(aos_attribute_t* attribute){
  /*char* string = getAttributeValueAsString(attribute);*/
  /*return strtol(string, NULL, 10);*/
  return *((int*) attribute->data);
}

void* getAttributeValueAsVoidPointer(aos_attribute_t* attribute){
  /*char* string = getAttributeValueAsString(attribute);*/
  /*return strtol(string, NULL, 10);*/
  return (void*) attribute->data;
}

/* AOS control functions */

/**
 * Initializes the AOS and creates thread-local storage key
 */
EES_RC AOS_Init (void){
  _is_threading = 0;
  pthread_once(&_aos_key_once, aos_make_key);

  if((_global_state = calloc(1, sizeof(aos_state_t)))){
    _global_state->last_context = NULL;
    if((_global_state->current_storage = calloc(1, sizeof(aos_storage_t)))){
      _global_state->current_storage->list_contexts = NULL;
      _global_state->saved_storage = _global_state->current_storage;
      _is_initialized = 1;
      return EES_SUCCESS;
    }
  }

  return EES_FAILURE;
}

/**
 * Creates thread-local storage key
 */
void aos_make_key(void){
  pthread_key_create(&_aos_key, aos_free_key);
}

/**
 * Frees thread-local storage key
 */
void aos_free_key(void* state){
  pid_t             tid;
  aos_state_t*      tmp_state = (aos_state_t*) state;
  tid = syscall(SYS_gettid);
  EEF_log(LOG_DEBUG, "Killing thread %i\n", tid);
  aos_free_state(tmp_state);
}

/**
 * Frees thread-local storage for this thread
 */
EES_RC AOS_Clean(void){
  pid_t             tid;
  tid = syscall(SYS_gettid);
  EEF_log(LOG_DEBUG, "Killing thread %i\n", tid);
  if(_is_threading){
    aos_free_state(aos_get_state());
  }
  return EES_SUCCESS;
}

/**
 * Terminates AOS, deleting the global storage key
 */
EES_RC AOS_Term (void){
  _is_threading = 0;
  aos_free_state(_global_state);
  pthread_setspecific(_aos_key, NULL);
  pthread_key_delete(_aos_key);
  _is_initialized = 0;
  return EES_SUCCESS;
}

/**
 * Dumps out all the data present in the AOS, as seen from the current thread's perspective
 */
EES_RC aos_dump_argslist (void) {
  aos_context_t   *context         = NULL;
  aos_attribute_t *attribute       = NULL;
  char            *attribute_name  = NULL;
  char            *attribute_value = NULL;

  EEF_log(LOG_DEBUG, "Dumping aos");

  rewindContexts(NULL);

  while((context = getNextContext(ANY, NULL)) != NULL){
    EEF_log(LOG_DEBUG, "Context %p class: %i\n", context, context->context_class);
    rewindAttributes(context);
    while((attribute = getNextAttribute(context)) != NULL){
      attribute_name = getAttributeId(attribute);
      attribute_value = getAttributeValueAsString(attribute);
      if(attribute_name && attribute_value){
        if(attribute->needs_free){
          EEF_log(LOG_DEBUG, "\t%s=%s\n", attribute_name, attribute_value);
        } else {
          EEF_log(LOG_DEBUG, "\t%s=%i\n", attribute_name, attribute_value);
        }
      }
    }
  }

  return EES_SUCCESS;
}

