#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>

#include "ees_eics.h"
#include "config.h"
#include "xacml.h"
#include "xacml_server.h"
#include "eef_library.h"
#include "cgul/log_to_file/log_to_file.h"
#include "cgul_patched/network/net_common.h"

#define _MULTI_THREADED
#define BACKLOG 100

char* config_file_path;                 /* config file path */
xacml_server_t server;                  /* xacml server object */
int signal_is_being_handled = 0;        /* signal handler state/lock */

void signal_handler(int sig);           /* main signal handler */
EES_RC process_xacml(int port);         /* creates and starts new xacml server instance */
EES_RC ees_start(void);                 /* initializes EEF */
EES_RC ees_stop(void);                  /* terminates EEF */
EES_RC ees_loop(int port);              /* main loop - calls process_xacml and pauses */

/**
 * Prepares logging, initializes EEF and puts the EEF in threading mode
 */
EES_RC ees_start(void){
    /* set log options */
/*  if(cgul_openlog("/var/log/ees/ees_log") != 0){
	return EES_FAILURE;
    }*/
    cgul_set_ident(BINARY_NAME);
/*  cgul_set_log_options(CGUL_SYSLOG | CGUL_LOGFILE | CGUL_STDERR);*/
    cgul_set_log_options(CGUL_SYSLOG);

    /* reuse this part when implementing SIGHUP */
    cgul_log(LOG_INFO, "Parsing config file %s\n", config_file_path);
    if(EEF_init(config_file_path, cgul_log) == EES_FAILURE){
    /* at this point, EEF_log can/will be a reference to the second argument passed to EEF_init */
	EEF_log(LOG_ERR, "Failed to parse config file (%s), exiting...", config_file_path);
	return EES_FAILURE;
    }

    EEF_startThreading();
    EEF_log(LOG_NOTICE, "EES Ready.\n");

    return EES_SUCCESS;
}

/**
 * Main EES loop. Starts xacml server and pauses.
 */
EES_RC ees_loop(int port){
  while(1){ /* this is to prevent falling out of the loop when a signal gets handled */
    signal_is_being_handled = 0;
    if(process_xacml(port) == EES_SUCCESS){
      printf("Paused.\n");
      pause();
    } else {
      EEF_log(LOG_ERR, "Unable to start xacml server\n");
      ees_stop();
      exit(EXIT_FAILURE);
    }
  }
  return EES_SUCCESS;
}

/**
 * Terminates EES
 */
EES_RC ees_stop(void){
  xacml_server_destroy(server);
  AOS_Clean();
  if(EEF_term() == EES_SUCCESS){
    EEF_log(LOG_NOTICE, "EEF terminated!\n");
  } else {
    EEF_log(LOG_NOTICE, "Error while terminating EEF!\n");
  }
  fflush(stdout);
  fflush(stderr);
  EEF_log(LOG_NOTICE, "EES Stopped.\n");  
  cgul_closelog();
  return EES_SUCCESS;
}

/**
 * Creates and starts xacml server.
 */
EES_RC process_xacml(int port){
  int fd;
  /*shutdown(fd, 0);*/
  if((fd = createAndSetUpATCPServerSocket(port, BACKLOG)) > -1){
    xacml_init(); /* if you look at the code, you'll see this function is empty */
    if(xacml_server_init(&server, ees_xacml_authorize, NULL) == XACML_RESULT_SUCCESS){
      xacml_server_set_fd(server, fd);
      xacml_server_start(server);
      return EES_SUCCESS;
    }
  }
  return EES_FAILURE;
}

/**
 * Handles incoming signals
 */
void signal_handler(int sig){
  printf("Got signal\n");
  if(signal_is_being_handled == 0){
    signal_is_being_handled = 1;
    switch(sig){
      case SIGHUP:
        EEF_log(LOG_NOTICE, "Got SIGHUP - restart \n");
        ees_stop();
        if(ees_start() == EES_FAILURE){
          EEF_log(LOG_ERR, "Unable to restart EES. Exiting...\n");
          exit(EXIT_FAILURE);
        }
        break;
      case SIGTERM:
        EEF_log(LOG_NOTICE, "Got SIGTERM - shutting down\n");
        ees_stop();
        exit(EXIT_SUCCESS);
        break;
    }
  }
}

/**
 * Main
 */
int main (int argc, char* argv[]){
  struct sigaction sa;
  int port;

	if(argc > 1){
		config_file_path = argv[1];
	} else {
		config_file_path = EES_CONFIG_FILE;
	}

  /* catch hangup and term signals */
  sa.sa_handler = signal_handler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART;

  if(sigaction(SIGHUP, &sa, NULL) != 0){
    EEF_log(LOG_ERR, "Unable to add signal handler for SIGHUP\n");
  }
  if(sigaction(SIGTERM, &sa, NULL) != 0){
    EEF_log(LOG_ERR, "Unable to add signal handler for SIGHUP\n");
  }

#if ENABLE_DEBUG
    cgul_set_log_options(cgul_get_log_options() | CGUL_STDDEBUG);
#endif
  if(ees_start() == EES_SUCCESS){
#if !ENABLE_DEBUG
    EEF_daemonize();
#endif
    printf("port=%d\n",EEF_get_port);
    if ( (port=EEF_get_port()) == -1)	{
	EEF_log(LOG_ERR, "Port is not set or invalid.\n");
	return EXIT_FAILURE;
    }

    ees_loop(port);
    return EXIT_SUCCESS;
  }

  return EXIT_FAILURE;

}

